#include "sock.h"

int FD_MAX;

/*
 * return a file descriptor for a new connection.
 * Return value -1 means error occurred.
 *
 * sock is a file descriptor for a server socket.
 * The (connecting) client's address is copied into the string pointed to by
 * ip_addr (if ip_addr is not NULL), and the port from which the client is
 * connecting is stored in the integer pointed to by port (if ip_addr and port 
 * is not NULL).
 */
int acceptConnection(int sock, char *ip_addr, int *port)
{
	int fd;
	struct sockaddr addr;
	struct sockaddr_in sin;
	int len = sizeof (struct sockaddr);

	/* accept a connection */
	fd = accept(sock, &addr, &len);
	if (fd < 0) {
		perror("acceptConnection: accept");
		return -1;
	}

	/* get the IP address of the client */
	if (ip_addr) {
		struct hostent *from = NULL;
		memmove(&sin, &addr, len);
		strcpy(ip_addr, inet_ntoa(sin.sin_addr));

		from = gethostbyaddr((char *) &sin.sin_addr,
		                     sizeof(sin.sin_addr), AF_INET);
		if (from) {
			strcpy(ip_addr, from->h_name);
		}
		
		if (port)
			*port = ntohs(sin.sin_port);
	}

	/* set the descriptor so that it is in no blocking mode */
/*	if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1) {
		perror("acceptConnection: fcntl");
		return -1;
	}
*/

//	ioctl(sock, FIONBIO, &i);

	/* connection successful, return the file descriptor */
	return fd;
}

/*
 * Add a Socket to the slist, with descriptor fd.
 */
void addSocket(SocketList *slist, int fd)
{
	assert(slist->n < CONN_MAX); /* CONN_MAX is the max size of the array */

	slist->sock[slist->n].fd = fd;
	slist->sock[slist->n].n = 0;
	slist->n++;
}

/*
 * Return true if there is any pending data on the socket described by sock.
 */
bool available(int sock)
{
	fd_set read_set;
	struct timeval tv;

	/* this will make select return immediately */
	tv.tv_sec = 0;
	tv.tv_usec = 0;

	/* add sock to the read_set */
	FD_ZERO(&read_set);
	FD_SET(sock, &read_set);

	/* check for pending data */
	if (select(FD_MAX, &read_set, NULL, NULL, &tv) > 0 &&
	    FD_ISSET(sock, &read_set))
		return true;
	else
		return false;
}

/*
 * Check each fd in slist for bytes available to be read. polltime is the time
 * in microseconds to poll the sockets.
 * Any pending input is read into the buf for the Socket. n (in Sock) stores 
 * the number bytes read, -1 on EOF.
 */
void checkIO(SocketList *slist, long polltime)
{
	int i;
	fd_set read_set;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = polltime;

	/* add all the file descriptors to the read_set */
	FD_ZERO(&read_set);
	for (i = 0; i < slist->n; i++) {
		FD_SET(slist->sock[i].fd, &read_set);
	}

	/* check the status of the file descriptors */
	if (select(FD_MAX, &read_set, NULL, NULL, &tv) > 0) {

		/* read BUF_SIZE bytes for each file descriptor which has
		 * pending data to be read.
		 */
		for (i = 0; i < slist->n; i++) {
			slist->sock[i].n = 0;

			/* read input to buffer */
			if (FD_ISSET(slist->sock[i].fd, &read_set)) {

				/* Windows doesn't like read() for sockets */
//				if (slist->sock[i].fd == 0) /* stdin */
//					slist->sock[i].n = read(slist->sock[i].fd,
//					                        slist->sock[i].buf,
//					                        BUF_SIZE);
//				else

				slist->sock[i].n = recv(slist->sock[i].fd,
				                        slist->sock[i].buf,
				                        BUF_SIZE,
				                        0);
				
				if (slist->sock[i].n <= 0)
					slist->sock[i].n = -1; /* close socket */
			}
		}
	}
}

/*
 * Check if the server socket sock has any waiting connections.
 */
bool checkServer(int sockfd)
{
	fd_set read_set;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 75000L;

	FD_ZERO(&read_set);
	FD_SET(sockfd, &read_set);

	if (select(FD_MAX, &read_set, NULL, NULL, &tv) < 0) {
		perror("checkServer: select");
		return false;
	}

	if (FD_ISSET(sockfd, &read_set))
		return true;
	else
		return false;
}

/*
 * Connect to the host address, port port.
 * Return value is a file descriptor for the socket.
 * Return value -1 indicates error.
 */
int initConnection(char *address, unsigned short port)
{
	int sock;
	struct sockaddr_in sin;
	unsigned long addr;
	struct hostent *h;

	/* set the address */
	/* by host name */
	if ((h = gethostbyname(address)) != NULL) {
		sin.sin_family = h->h_addrtype;
		memcpy(&addr, h->h_addr, h->h_length);/* hostent is in netdb.h */
	}
	/* by IP number */
	else if ((addr = inet_addr(address)) != -1) {
		sin.sin_family = AF_INET;
	}
	/* failed by IP number and host name */
	else {
		fprintf(stderr, "Host not found: %s\n", address);
		// perror("initConnection");
		return -1;
	}
	sin.sin_addr.s_addr = addr;  /* store the address */
//	sin.sin_addr = addr;         /* alternative ?     */
	sin.sin_port = htons(port);  /* port              */

	/* create socket */
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("initConnection: socket");
		return -1;
	}

	/* connect to the port */
	if (connect(sock, (struct sockaddr *) &sin, sizeof (sin)) == -1) {
		// perror("initConnection: connect");
#ifdef WIN32
		closesocket(sock);
#else
		close(sock);
#endif
		perror("connect");
		return -1;
	}

	/* return the file descriptor for the socket */
	return sock;
}

/*
 * Return a server socket descriptor.
 * Return value -1 indicates error.
 */
int initServer(unsigned short port)
{

	struct sockaddr_in addr;
	int sock;
#ifdef WIN32
	char i = 1;
#else
	int i = 1;
#endif

	/* set port number, and other stuff */
	addr.sin_port = htons(port);       /* duh...                       */
	addr.sin_family = AF_INET;         /* ARPA Internet protocol       */
	addr.sin_addr.s_addr = INADDR_ANY; /* accept any incoming messages */

	/* create socket */
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("initServer: socket");
		return -1;
	}

	/* allow connections to occur more than once */
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (i));

	/* bind to address */
	if (bind(sock, (struct sockaddr *) &addr, sizeof (addr)) == -1) {
		perror("initServer: bind");
		return -1;
	}

	/* be willing to listen for connections */
	if (listen(sock, 5) == -1) {
		perror("initServer: listen");
		return -1;
	}

//	ioctl(sock, FIONBIO, &i);

	return sock;
}

/*
 * Initialise the SocketList slist.
 */
void initSocketList(SocketList *slist)
{
#ifndef WIN32
	FD_MAX = getdtablesize();  /* this is set in socketInit() for WIN32 */
#endif
	assert(FD_MAX > 0);
	slist->n = 0;
}

/*
 * Check slist for any data in the buffers.
 * If there is data then processInput(int fd, char *buf, int n) is called
 * with the following parameters:
 *   fd  - the socket descriptor for a socket, which has input, in slist
 *   buf - the input buffer for the socket
 *   n   - number of bytes in buf
 * processInput must return a nonzero value if the socket fd is to be closed.
 */
void processData(SocketList *slist, int (*processInput)(int, char *, int))
{
	int i;

	for (i = 0; i < slist->n; i++) {

		/* if bytes were read (n > 0) then call processInput */
		if (slist->sock[i].n > 0) {
			if ((*processInput)(slist->sock[i].fd,
			                    slist->sock[i].buf,
			                    slist->sock[i].n) != 0)
				slist->sock[i].n = -1; /* close socket */
			else
				slist->sock[i].n = 0;
		}

		/* If we have reached EOF or the connection has been closed,
		 * then close the socket.
		 */
		if (slist->sock[i].n == -1) {
			printf("[Connection closed]\n");

#ifdef WIN32
			closesocket(slist->sock[i].fd);
#else
			close(slist->sock[i].fd);
#endif

			/* remove the socket we just closed */
			removeSocket(slist, slist->sock[i].fd);
			i--; /* process the Socket moved by removeSocket */
		}
	}
}

/*
 * Remove the socket from slist which has the file descriptor fd.
 *
 * Return value is true on success, false if the socket fd was not found in 
 * slist. All Sockets occuring before the Socket fd in slist remain where 
 * they are.
 */
bool removeSocket(SocketList *slist, int fd)
{
	int i;

	for (i = 0; i < slist->n; i++) {
		if (slist->sock[i].fd == fd)
			break;
	}

	/* if no match was found, return -1 */
	if (i == slist->n)
		return false;

	/* move the last Socket in slist to fill the Socket being removed */
	if (slist->n > 1) {
		slist->sock[i] = slist->sock[slist->n - 1];
	}

	/* One less Socket */
	slist->n--;

	return true;
}

/* 
 * Send a string message to a socket
 */
void sendmesg(int fd, char *msg)
{
	send(fd, msg, strlen(msg), 0);
}

/*
 * Initialize the socket library. Call before using any functions.
 */
bool socketInit(void)
{

#ifdef WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 0 );
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		return false;
	}

	FD_MAX = 128;	/* not used anyway by windows, but used for SocketLists */
#else
	FD_MAX = getdtablesize();
#endif

	return true;
}

/*
 * Shut down the socket library. Call after finishing.
 */
bool socketShutdown(void)
{

#ifdef WIN32
	if (WSACleanup() != 0)
		return false;
	else
#endif
		return true;
}

/*
 * Open a test connection to host:port. If successful, true is returned.
 * If the connection cannot be established, false is returned.
 */
bool testConnection(char *host, unsigned short port)
{
	int fd;

//	fprintf(stderr, "Checking %s:%d...", host, port);

	fd = initConnection(host, port);

	if (fd == -1) {
		/* initConnection prints an error message */

		return false;
	}
	else {
//		fprintf(stderr, "OK\n");
#ifdef WIN32
		closesocket(fd);
#else
		close(fd);
#endif
		return true;
	}
}
