/*
 * This is a general purpose socket library which will work under both
 * Windows 95/NT and UNIX / Linux.
 *
 * Copyright (C) 1998  ztank
 *
 * Before using any of these functions, call
 *   socketInit();
 * When finished, call
 *   socketShutdown();         (only required for Windows)
 *
 * To initialise a server use
 *   int serversockfd = initServer(port);
 * initServer() returns a socket fd, or -1 on error.
 *
 * To accept a connection on the server use
 *   if (checkServer(serversockfd)) {
 *     char ip_addr[100];
 *     int port;
 *     int sockfd = acceptConnection(serversockfd, ip_addr, &port);
 *     ...
 *   }
 * checkServer() returns true is there are any waiting clients to connect
 * acceptConnection() returns a socket fd for the client, or -1 on error;
 *   if there are no clients available acceptConnection() blocks until one
 *   becomes available. If ip_addr and port are not NULL, the connecting
 *   client's address (host name if possible) and port are stored here.
 *
 *
 * To obtain data waiting on existing sockets create a SocketList with
 *   SocketList slist;
 *   initSocketList(&slist);
 * To add sockets to be checked for data, use
 *   addSocket(&slist, sockfd);  // sockfd = 0 for stdin
 * To remove sockets to be checked for data, use
 *   removeSocket(&slist, sockfd);
 * Once this is done, to check these sockets for data use
 *   checkIO(&slist, poll); // reads data into Socket buffers after polling
 *                          // sockets for "poll" microseconds
 *   processData(&slist, processInput);
 * The function processInput() is called by processData() for each socket fd 
 * with data in its buffer. processInput() is a user defined function:
 *   int processInput(int fd, char *buf, int n)
 *     fd - socket fd
 *     buf - data which was read in
 *     n - number of bytes in buf
 *     return value - nonzero if this socket is to be closed, 0 otherwise
 *
 * An alternative way to accept data is to call available() with a socket
 * descriptor. available() returns true if there is any data waiting on the
 * socket:
 *   if (available(sock))
 *     ... recv(sock, ...) ...
 *
 * To connect to a server, use
 *   int sockfd = initConnection(address, port);
 * address can be a hostname or an IP address. initConnection() returns a
 * socket fd, or -1 on error.
 *
 * As a diagnostic tool a function
 *   testConnection(host, port);
 * is provided to check to see if a connection can be made. true is returned
 * is a connection can be successfully made.
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef WIN32
#include <windows.h>
#include <time.h>
#include <winsock.h>

#else
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#endif	/* WIN32 */

#define CONN_MAX 10
#define BUF_SIZE 10240
#define true    1
#define false   0
typedef int bool;

typedef struct Socket {
	int fd;
	char buf[BUF_SIZE];
	int n;
} Socket;

/* A list of Sockets */
typedef struct SocketList {
	Socket sock[CONN_MAX];
	int n;
} SocketList;

int acceptConnection(int sock, char *ip_addr, int *port);
bool available(int sock);
bool checkServer(int sockfd);
int initConnection(char *address, unsigned short port);
int initServer(unsigned short port);
void sendmesg(int fd, char *msg);
bool socketInit(void);
bool socketShutdown(void);
bool testConnection(char *host, unsigned short port);

/* SocketList related functions */
void addSocket(SocketList *slist, int fd);
void checkIO(SocketList *slist, long polltime);
void initSocketList(SocketList *slist);
void processData(SocketList *slist, int (*processInput)(int, char *, int));
bool removeSocket(SocketList *slist, int fd);
